package ru.dbotthepony.kbox2d.dynamics.joint

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.api.B2SolverData
import ru.dbotthepony.kbox2d.api.b2Mul
import ru.dbotthepony.kstarbound.math.*

class WeldJoint(def: WeldJointDef) : AbstractJoint(def) {
	var stiffness: Double = def.stiffness
	var damping: Double = def.damping
	private var bias: Double = 0.0

	// Solver shared
	val localAnchorA = def.localAnchorA
	val localAnchorB = def.localAnchorB
	val referenceAngle = def.referenceAngle
	private var gamma: Double = 0.0
	private var impulse: Vector3d = Vector3d()

	// Solver temp
	private var indexA: Int = 0
	private var indexB: Int = 0
	private var rA: Vector2d = Vector2d.ZERO
	private var rB: Vector2d = Vector2d.ZERO
	private var localCenterA: Vector2d = Vector2d.ZERO
	private var localCenterB: Vector2d = Vector2d.ZERO
	private var invMassA: Double = 0.0
	private var invMassB: Double = 0.0
	private var invIA: Double = 0.0
	private var invIB: Double = 0.0
	private var mass: MutableMatrix3d = MutableMatrix3d().also { it.zero() }

	override fun initVelocityConstraints(data: B2SolverData) {
		this.indexA = this.bodyA.islandIndex
		this.indexB = this.bodyB.islandIndex
		this.localCenterA = this.bodyA.sweep.localCenter
		this.localCenterB = this.bodyB.sweep.localCenter
		this.invMassA = this.bodyA.invMass
		this.invMassB = this.bodyB.invMass
		this.invIA = this.bodyA.invI
		this.invIB = this.bodyB.invI

		val aA = data.positions[this.indexA].a
		var vA = data.velocities[this.indexA].v
		var wA = data.velocities[this.indexA].w

		val aB = data.positions[this.indexB].a
		var vB = data.velocities[this.indexB].v
		var wB = data.velocities[this.indexB].w

		val qA = Rotation(aA)
		val qB = Rotation(aB)

		this.rA = b2Mul(qA, this.localAnchorA - this.localCenterA)
		this.rB = b2Mul(qB, this.localAnchorB - this.localCenterB)

		// J = [-I -r1_skew I r2_skew]
		//     [ 0       -1 0       1]
		// r_skew = [-ry; rx]

		// Matlab
		// K = [ mA+r1y^2*iA+mB+r2y^2*iB,  -r1y*iA*r1x-r2y*iB*r2x,          -r1y*iA-r2y*iB]
		//     [  -r1y*iA*r1x-r2y*iB*r2x, mA+r1x^2*iA+mB+r2x^2*iB,           r1x*iA+r2x*iB]
		//     [          -r1y*iA-r2y*iB,           r1x*iA+r2x*iB,                   iA+iB]

		val mA = this.invMassA
		val mB = this.invMassB
		val iA = this.invIA
		val iB = this.invIB

		val K = MutableMatrix3d()
		K.m00 = mA + mB + this.rA.y * this.rA.y * iA + this.rB.y * this.rB.y * iB
		K.m01 = -this.rA.y * this.rA.x * iA - this.rB.y * this.rB.x * iB
		K.m02 = -this.rA.y * iA - this.rB.y * iB

		K.m10 = K.m01
		K.m11 = mA + mB + this.rA.x * this.rA.x * iA + this.rB.x * this.rB.x * iB
		K.m12 = this.rA.x * iA + this.rB.x * iB

		K.m20 = K.m02
		K.m21 = K.m12
		K.m22 = iA + iB

		if (this.stiffness > 0.0) {
			this.mass = K.getInverse2().asMutableMatrix()

			var invM = iA + iB

			val C = aB - aA - this.referenceAngle

			// Damping coefficient
			val d = this.damping

			// Spring stiffness
			val k = this.stiffness

			// magic formulas
			val h = data.step.dt
			this.gamma = h * (d + h * k)
			this.gamma = if (this.gamma != 0.0) 1.0 / this.gamma else 0.0
			this.bias = C * h * k * this.gamma

			invM += this.gamma
			this.mass.m22 = if (invM != 0.0) 1.0 / invM else 0.0
		} else if (K.m22 == 0.0) {
			this.mass = K.getInverse2().asMutableMatrix()
			this.gamma = 0.0
			this.bias = 0.0
		} else {
			this.mass = K.getInverse().asMutableMatrix()
			this.gamma = 0.0
			this.bias = 0.0
		}

		if (data.step.warmStarting) {
			// Scale impulses to support a variable time step.
			this.impulse *= data.step.dtRatio

			val P = Vector2d(this.impulse.x, this.impulse.y)

			vA -= mA * P
			wA -= iA * (b2Cross(this.rA, P) + this.impulse.z)

			vB += mB * P
			wB += iB * (b2Cross(this.rB, P) + this.impulse.z)
		} else {
			this.impulse = Vector3d()
		}

		data.velocities[this.indexA].v = vA
		data.velocities[this.indexA].w = wA
		data.velocities[this.indexB].v = vB
		data.velocities[this.indexB].w = wB
	}

	override fun solveVelocityConstraints(data: B2SolverData) {
		var vA = data.velocities[this.indexA].v
		var wA = data.velocities[this.indexA].w
		var vB = data.velocities[this.indexB].v
		var wB = data.velocities[this.indexB].w

		val mA = this.invMassA
		val mB = this.invMassB
		val iA = this.invIA
		val iB = this.invIB

		if (this.stiffness > 0.0) {
			val Cdot2 = wB - wA

			val impulse2 = -this.mass.m22 * (Cdot2 + this.bias + this.gamma * this.impulse.z)
			this.impulse += Vector3d(z = impulse2)

			wA -= iA * impulse2
			wB += iB * impulse2

			val Cdot1 = vB + b2Cross(wB, this.rB) - vA - b2Cross(wA, this.rA)

			val impulse1 = -b2Mul22(this.mass, Cdot1)
			this.impulse += Vector3d(impulse1.x, impulse1.y)

			val P = impulse1

			vA -= mA * P
			wA -= iA * b2Cross(this.rA, P)

			vB += mB * P
			wB += iB * b2Cross(this.rB, P)
		} else {
			val Cdot1 = vB + b2Cross(wB, this.rB) - vA - b2Cross(wA, this.rA)
			val Cdot2 = wB - wA
			val Cdot = Vector3d(Cdot1.x, Cdot1.y, Cdot2)

			val impulse = -b2Mul(this.mass, Cdot)
			this.impulse += impulse

			val P = Vector2d(impulse.x, impulse.y)

			vA -= mA * P
			wA -= iA * (b2Cross(this.rA, P) + impulse.z)

			vB += mB * P
			wB += iB * (b2Cross(this.rB, P) + impulse.z)
		}

		data.velocities[this.indexA].v = vA
		data.velocities[this.indexA].w = wA
		data.velocities[this.indexB].v = vB
		data.velocities[this.indexB].w = wB
	}

	override fun solvePositionConstraints(data: B2SolverData): Boolean {
		var cA = data.positions[this.indexA].c
		var aA = data.positions[this.indexA].a
		var cB = data.positions[this.indexB].c
		var aB = data.positions[this.indexB].a

		val qA = Rotation(aA)
		val qB = Rotation(aB)

		val mA = this.invMassA
		val mB = this.invMassB
		val iA = this.invIA
		val iB = this.invIB

		val rA = b2Mul(qA, this.localAnchorA - this.localCenterA)
		val rB = b2Mul(qB, this.localAnchorB - this.localCenterB)

		val positionError: Double
		val angularError: Double

		val K = MutableMatrix3d()
		K.m00 = mA + mB + rA.y * rA.y * iA + rB.y * rB.y * iB
		K.m01 = -rA.y * rA.x * iA - rB.y * rB.x * iB
		K.m02 = -rA.y * iA - rB.y * iB
		K.m10 = K.m01
		K.m11 = mA + mB + rA.x * rA.x * iA + rB.x * rB.x * iB
		K.m12 = rA.x * iA + rB.x * iB
		K.m20 = K.m02
		K.m21 = K.m12
		K.m22 = iA + iB

		if (this.stiffness > 0.0f) {
			val C1 = cB + rB - cA - rA

			positionError = C1.length
			angularError = 0.0

			val P = -K.solve(C1)

			cA -= mA * P
			aA -= iA * b2Cross(rA, P)

			cB += mB * P
			aB += iB * b2Cross(rB, P)
		} else {
			val C1 = cB + rB - cA - rA
			val C2 = aB - aA - this.referenceAngle

			positionError = C1.length
			angularError = b2Abs(C2)

			val C = Vector3d(C1.x, C1.y, C2)

			val impulse: Vector3d
			if (K.m22 > 0.0) {
				impulse = -K.solve(C)
			} else {
				val impulse2 = -K.solve(C1)
				impulse = Vector3d(impulse2.x, impulse2.y)
			}

			val P = Vector2d(impulse.x, impulse.y)

			cA -= mA * P
			aA -= iA * (b2Cross(rA, P) + impulse.z)

			cB += mB * P
			aB += iB * (b2Cross(rB, P) + impulse.z)
		}

		data.positions[this.indexA].c = cA
		data.positions[this.indexA].a = aA
		data.positions[this.indexB].c = cB
		data.positions[this.indexB].a = aB

		return positionError <= b2_linearSlop && angularError <= b2_angularSlop
	}

	override val anchorA: Vector2d
		get() = bodyA.getWorldPoint(localAnchorA)
	override val anchorB: Vector2d
		get() = bodyB.getWorldPoint(localAnchorB)

	override fun getReactionForce(inv_dt: Double): Vector2d {
		return Vector2d(impulse.x * inv_dt, impulse.y * inv_dt)
	}

	override fun getReactionTorque(inv_dt: Double): Double {
		return inv_dt * impulse.z
	}
}
