package ru.dbotthepony.kbox2d.dynamics.joint

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.api.B2SolverData
import ru.dbotthepony.kbox2d.api.b2Cross
import ru.dbotthepony.kstarbound.math.Vector2d
import ru.dbotthepony.kstarbound.math.times
import ru.dbotthepony.kstarbound.util.Color

class WheelJoint(def: WheelJointDef) : AbstractJoint(def) {
	val localAnchorA: Vector2d = def.localAnchorA
	val localAnchorB: Vector2d = def.localAnchorB
	val localXAxisA: Vector2d = def.localAxisA
	val localYAxisA: Vector2d = b2Cross(1.0, localXAxisA)

	private var impulse: Double = 0.0
	private var motorImpulse: Double = 0.0
	private var springImpulse: Double = 0.0

	private var lowerImpulse: Double = 0.0
	private var upperImpulse: Double = 0.0
	private var translation: Double = 0.0

	// Solver temp
	private var indexA: Int = 0
	private var indexB: Int = 0
	private var localCenterA: Vector2d = Vector2d.ZERO
	private var localCenterB: Vector2d = Vector2d.ZERO
	private var invMassA: Double = 0.0
	private var invMassB: Double = 0.0
	private var invIA: Double = 0.0
	private var invIB: Double = 0.0

	private var ax: Vector2d = Vector2d.ZERO
	private var ay: Vector2d = Vector2d.ZERO

	private var sAx: Double = 0.0
	private var sBx: Double = 0.0
	private var sAy: Double = 0.0
	private var sBy: Double = 0.0

	private var mass: Double = 0.0
	private var motorMass: Double = 0.0
	private var axialMass: Double = 0.0
	private var springMass: Double = 0.0

	private var bias: Double = 0.0
	private var gamma: Double = 0.0

	override fun initVelocityConstraints(data: B2SolverData) {
		this.indexA = this.bodyA.islandIndex
		this.indexB = this.bodyB.islandIndex
		this.localCenterA = this.bodyA.sweep.localCenter
		this.localCenterB = this.bodyB.sweep.localCenter
		this.invMassA = this.bodyA.invMass
		this.invMassB = this.bodyB.invMass
		this.invIA = this.bodyA.invI
		this.invIB = this.bodyB.invI

		val mA = this.invMassA
		val mB = this.invMassB
		val iA = this.invIA
		val iB = this.invIB

		val cA = data.positions[this.indexA].c
		val aA = data.positions[this.indexA].a
		var vA = data.velocities[this.indexA].v
		var wA = data.velocities[this.indexA].w

		val cB = data.positions[this.indexB].c
		val aB = data.positions[this.indexB].a
		var vB = data.velocities[this.indexB].v
		var wB = data.velocities[this.indexB].w

		val qA = Rotation(aA)
		val qB = Rotation(aB)

		// Compute the effective masses.
		val rA = b2Mul(qA, this.localAnchorA - this.localCenterA)
		val rB = b2Mul(qB, this.localAnchorB - this.localCenterB)
		val d = cB + rB - cA - rA

		// Point to line constraint
		run {
			this.ay = b2Mul(qA, this.localYAxisA)
			this.sAy = b2Cross(d + rA, this.ay)
			this.sBy = b2Cross(rB, this.ay)

			this.mass = mA + mB + iA * this.sAy * this.sAy + iB * this.sBy * this.sBy

			if (this.mass > 0.0) {
				this.mass = 1.0 / this.mass
			}
		}

		// Spring constraint
		this.ax = b2Mul(qA, this.localXAxisA)
		this.sAx = b2Cross(d + rA, this.ax)
		this.sBx = b2Cross(rB, this.ax)

		val invMass = mA + mB + iA * this.sAx * this.sAx + iB * this.sBx * this.sBx

		if (invMass > 0.0) {
			this.axialMass = 1.0 / invMass
		} else {
			this.axialMass = 0.0
		}

		this.springMass = 0.0
		this.bias = 0.0
		this.gamma = 0.0

		if (this.stiffness > 0.0f && invMass > 0.0f) {
			this.springMass = 1.0f / invMass

			val C = b2Dot(d, this.ax)

			// magic formulas
			val h = data.step.dt
			this.gamma = h * (this.damping + h * this.stiffness)
			if (this.gamma > 0.0f) {
				this.gamma = 1.0f / this.gamma
			}

			this.bias = C * h * this.stiffness * this.gamma

			this.springMass = invMass + this.gamma
			if (this.springMass > 0.0) {
				this.springMass = 1.0 / this.springMass
			}
		} else {
			this.springImpulse = 0.0
		}

		if (this.enableLimit) {
			this.translation = b2Dot(this.ax, d)
		} else {
			this.lowerImpulse = 0.0
			this.upperImpulse = 0.0
		}

		if (this.enableMotor) {
			this.motorMass = iA + iB

			if (this.motorMass > 0.0f) {
				this.motorMass = 1.0f / this.motorMass
			}
		} else {
			this.motorMass = 0.0
			this.motorImpulse = 0.0
		}

		if (data.step.warmStarting) {
			// Account for variable time step.
			this.impulse *= data.step.dtRatio
			this.springImpulse *= data.step.dtRatio
			this.motorImpulse *= data.step.dtRatio

			val axialImpulse = this.springImpulse + this.lowerImpulse - this.upperImpulse
			val P = this.impulse * this.ay + axialImpulse * this.ax
			val LA = this.impulse * this.sAy + axialImpulse * this.sAx + this.motorImpulse
			val LB = this.impulse * this.sBy + axialImpulse * this.sBx + this.motorImpulse

			vA -= this.invMassA * P
			wA -= this.invIA * LA

			vB += this.invMassB * P
			wB += this.invIB * LB
		} else {
			this.impulse = 0.0
			this.springImpulse = 0.0
			this.motorImpulse = 0.0
			this.lowerImpulse = 0.0
			this.upperImpulse = 0.0
		}

		data.velocities[this.indexA].v = vA
		data.velocities[this.indexA].w = wA
		data.velocities[this.indexB].v = vB
		data.velocities[this.indexB].w = wB
	}

	override fun solveVelocityConstraints(data: B2SolverData) {
		val mA = this.invMassA
		val mB = this.invMassB
		val iA = this.invIA
		val iB = this.invIB

		var vA = data.velocities[this.indexA].v
		var wA = data.velocities[this.indexA].w
		var vB = data.velocities[this.indexB].v
		var wB = data.velocities[this.indexB].w

		// Solve spring constraint
		run {
			val Cdot = b2Dot(this.ax, vB - vA) + this.sBx * wB - this.sAx * wA
			val impulse = -this.springMass * (Cdot + this.bias + this.gamma * this.springImpulse)
			this.springImpulse += impulse

			val P = impulse * this.ax
			val LA = impulse * this.sAx
			val LB = impulse * this.sBx

			vA -= mA * P
			wA -= iA * LA

			vB += mB * P
			wB += iB * LB
		}

		// Solve rotational motor constraint
		run {
			val Cdot = wB - wA - this.motorSpeed
			var impulse = -this.motorMass * Cdot

			val oldImpulse = this.motorImpulse
			val maxImpulse = data.step.dt * this.maxMotorTorque
			this.motorImpulse = b2Clamp(this.motorImpulse + impulse, -maxImpulse, maxImpulse)
			impulse = this.motorImpulse - oldImpulse

			wA -= iA * impulse
			wB += iB * impulse
		}

		if (this.enableLimit) {
			// Lower limit
			run {
				val C = this.translation - this.lowerTranslation
				val Cdot = b2Dot(this.ax, vB - vA) + this.sBx * wB - this.sAx * wA
				var impulse = -this.axialMass * (Cdot + b2Max(C, 0.0) * data.step.inv_dt)
				val oldImpulse = this.lowerImpulse
				this.lowerImpulse = b2Max(this.lowerImpulse + impulse, 0.0)
				impulse = this.lowerImpulse - oldImpulse

				val P = impulse * this.ax
				val LA = impulse * this.sAx
				val LB = impulse * this.sBx

				vA -= mA * P
				wA -= iA * LA
				vB += mB * P
				wB += iB * LB
			}

			// Upper limit
			// Note: signs are flipped to keep C positive when the constraint is satisfied.
			// This also keeps the impulse positive when the limit is active.
			run {
				val C = this.upperTranslation - this.translation
				val Cdot = b2Dot(this.ax, vA - vB) + this.sAx * wA - this.sBx * wB
				var impulse = -this.axialMass * (Cdot + b2Max(C, 0.0) * data.step.inv_dt)
				val oldImpulse = this.upperImpulse
				this.upperImpulse = b2Max(this.upperImpulse + impulse, 0.0)
				impulse = this.upperImpulse - oldImpulse

				val P = impulse * this.ax
				val LA = impulse * this.sAx
				val LB = impulse * this.sBx

				vA += mA * P
				wA += iA * LA
				vB -= mB * P
				wB -= iB * LB
			}
		}

		// Solve point to line constraint
		run {
			val Cdot = b2Dot(this.ay, vB - vA) + this.sBy * wB - this.sAy * wA
			val impulse = -this.mass * Cdot
			this.impulse += impulse

			val P = impulse * this.ay
			val LA = impulse * this.sAy
			val LB = impulse * this.sBy

			vA -= mA * P
			wA -= iA * LA

			vB += mB * P
			wB += iB * LB
		}

		data.velocities[this.indexA].v = vA
		data.velocities[this.indexA].w = wA
		data.velocities[this.indexB].v = vB
		data.velocities[this.indexB].w = wB
	}

	override fun solvePositionConstraints(data: B2SolverData): Boolean {
		var cA = data.positions[this.indexA].c
		var aA = data.positions[this.indexA].a
		var cB = data.positions[this.indexB].c
		var aB = data.positions[this.indexB].a

		var linearError = 0.0

		if (this.enableLimit) {
			val qA = Rotation(aA)
			val qB = Rotation(aB)

			val rA = b2Mul(qA, this.localAnchorA - this.localCenterA)
			val rB = b2Mul(qB, this.localAnchorB - this.localCenterB)
			val d = (cB - cA) + rB - rA

			val ax = b2Mul(qA, this.localXAxisA)
			val sAx = b2Cross(d + rA, this.ax)
			val sBx = b2Cross(rB, this.ax)

			var C = 0.0
			val translation = b2Dot(ax, d)
			if (b2Abs(this.upperTranslation - this.lowerTranslation) < 2.0 * b2_linearSlop) {
				C = translation
			} else if (translation <= this.lowerTranslation) {
				C = b2Min(translation - this.lowerTranslation, 0.0)
			} else if (translation >= this.upperTranslation) {
				C = b2Max(translation - this.upperTranslation, 0.0)
			}

			if (C != 0.0) {
				val invMass = this.invMassA + this.invMassB + this.invIA * sAx * sAx + this.invIB * sBx * sBx
				var impulse = 0.0
				if (invMass != 0.0) {
					impulse = -C / invMass
				}

				val P = impulse * ax
				val LA = impulse * sAx
				val LB = impulse * sBx

				cA -= this.invMassA * P
				aA -= this.invIA * LA
				cB += this.invMassB * P
				aB += this.invIB * LB

				linearError = b2Abs(C)
			}
		}

		// Solve perpendicular constraint
		run {
			val qA = Rotation(aA)
			val qB = Rotation(aB)

			val rA = b2Mul(qA, this.localAnchorA - this.localCenterA)
			val rB = b2Mul(qB, this.localAnchorB - this.localCenterB)
			val d = (cB - cA) + rB - rA

			val ay = b2Mul(qA, this.localYAxisA)

			val sAy = b2Cross(d + rA, ay)
			val sBy = b2Cross(rB, ay)

			val C = b2Dot(d, ay)

			val invMass = this.invMassA + this.invMassB + this.invIA * this.sAy * this.sAy + this.invIB * this.sBy * this.sBy

			var impulse = 0.0
			if (invMass != 0.0) {
				impulse = - C / invMass
			}

			val P = impulse * ay
			val LA = impulse * sAy
			val LB = impulse * sBy

			cA -= this.invMassA * P
			aA -= this.invIA * LA
			cB += this.invMassB * P
			aB += this.invIB * LB

			linearError = b2Max(linearError, b2Abs(C))
		}

		data.positions[this.indexA].c = cA
		data.positions[this.indexA].a = aA
		data.positions[this.indexB].c = cB
		data.positions[this.indexB].a = aB

		return linearError <= b2_linearSlop
	}

	override val anchorA: Vector2d
		get() = bodyA.getWorldPoint(localAnchorA)
	override val anchorB: Vector2d
		get() = bodyB.getWorldPoint(localAnchorB)

	override fun getReactionForce(inv_dt: Double): Vector2d {
		return inv_dt * (impulse * ay + (springImpulse + lowerImpulse - upperImpulse) * ax)
	}

	override fun getReactionTorque(inv_dt: Double): Double {
		return inv_dt * motorImpulse
	}

	val jointTranslation: Double get() {
		val bA = bodyA
		val bB = bodyB

		val pA = bA.getWorldPoint(localAnchorA)
		val pB = bB.getWorldPoint(localAnchorB)
		val d = pB - pA
		val axis = bA.getWorldVector(localXAxisA)

		return b2Dot(d, axis)
	}

	val jointLinearSpeed: Double get() {
		val bA = bodyA
		val bB = bodyB

		val rA = b2Mul(bA.xf.q, localAnchorA - bA.sweep.localCenter)
		val rB = b2Mul(bB.xf.q, localAnchorB - bB.sweep.localCenter)
		val p1 = bA.sweep.c + rA
		val p2 = bB.sweep.c + rB
		val d = p2 - p1
		val axis = b2Mul(bA.xf.q, localXAxisA)

		val vA = bA.linearVelocity
		val vB = bB.linearVelocity
		val wA = bA.angularVelocity
		val wB = bB.angularVelocity

		return b2Dot(d, b2Cross(wA, axis)) + b2Dot(axis, vB + b2Cross(wB, rB) - vA - b2Cross(wA, rA))
	}

	val jointAngle: Double get() {
		return bodyB.sweep.a - bodyA.sweep.a
	}

	val jointAngularSpeed: Double get() {
		return bodyB.angularVelocity - bodyA.angularVelocity
	}

	var lowerTranslation: Double = def.lowerTranslation
		private set
	var upperTranslation: Double = def.upperTranslation
		private set

	init {
		require(lowerTranslation <= upperTranslation) { "$lowerTranslation !<= $upperTranslation" }
	}

	fun setLimits(lower: Double, upper: Double) {
		require(lower <= upper) { "$lower !<= $upper" }

		if (lower != lowerTranslation || upper != upperTranslation) {
			bodyA.isAwake = true
			bodyB.isAwake = true
			lowerTranslation = lower
			upperTranslation = upper
			lowerImpulse = 0.0
			upperImpulse = 0.0
		}
	}

	var maxMotorTorque: Double = 0.0
		set(value) {
			if (field != value) {
				bodyA.isAwake = true
				bodyB.isAwake = true
				field = value
			}
		}

	fun getMotorTorque(inv_dt: Double): Double {
		return inv_dt * motorImpulse
	}

	var motorSpeed: Double = 0.0
		set(value) {
			if (field != value) {
				bodyA.isAwake = true
				bodyB.isAwake = true
				field = value
			}
		}

	var enableLimit: Boolean = def.enableLimit
		set(value) {
			if (field != value) {
				bodyA.isAwake = true
				bodyB.isAwake = true
				field = value
				lowerImpulse = 0.0
				upperImpulse = 0.0
			}
		}

	var enableMotor: Boolean = def.enableMotor
		set(value) {
			if (field != value) {
				bodyA.isAwake = true
				bodyB.isAwake = true
				field = value
			}
		}

	var stiffness: Double = def.stiffness
	var damping: Double = def.damping

	override fun draw(draw: IDebugDraw) {
		val xfA = bodyA.transform
		val xfB = bodyB.transform
		val pA = b2Mul(xfA, localAnchorA)
		val pB = b2Mul(xfB, localAnchorB)

		val axis = b2Mul(xfA.q, localXAxisA)

		draw.drawSegment(pA, pB, c5)

		if (enableLimit) {
			val lower = pA + lowerTranslation * axis
			val upper = pA + upperTranslation * axis
			val perp = b2Mul(xfA.q, localYAxisA)
			draw.drawSegment(lower, upper, c1)
			draw.drawSegment(lower - 0.5 * perp, lower + 0.5 * perp, c2)
			draw.drawSegment(upper - 0.5 * perp, upper + 0.5 * perp, c3)
		} else {
			draw.drawSegment(pA - 1.0 * axis, pA + 1.0 * axis, c1)
		}

		draw.drawPoint(pA, 5.0, c1)
		draw.drawPoint(pB, 5.0, c4)
	}

	companion object {
		private val c1 = Color(0.7f, 0.7f, 0.7f)
		private val c2 = Color(0.3f, 0.9f, 0.3f)
		private val c3 = Color(0.9f, 0.3f, 0.3f)
		private val c4 = Color(0.3f, 0.3f, 0.9f)
		private val c5 = Color(0.4f, 0.4f, 0.4f)
	}
}
