package ru.dbotthepony.kstarbound.client.gl

import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import org.lwjgl.opengl.GL46.*
import ru.dbotthepony.kstarbound.api.IStruct3f
import ru.dbotthepony.kstarbound.api.IStruct4f
import ru.dbotthepony.kstarbound.math.AbstractMatrix3f
import ru.dbotthepony.kstarbound.math.AbstractMatrix4f
import ru.dbotthepony.kstarbound.math.FloatMatrix
import java.util.*
import kotlin.collections.HashSet

class ShaderLinkException(reason: String) : RuntimeException(reason)

data class Uniform4f(val x: Float, val y: Float, val z: Float, val w: Float) : IStruct4f
data class Uniform3f(val x: Float, val y: Float, val z: Float) : IStruct3f

class GLUniformLocation(val program: GLShaderProgram, val name: String, val pointer: Int) {
	fun set(value: IStruct4f): GLUniformLocation {
		program.state.ensureSameThread()
		val (v0, v1, v2, v3) = value
		glProgramUniform4f(program.pointer, pointer, v0, v1, v2, v3)
		return this
	}

	fun set(value: IStruct3f): GLUniformLocation {
		program.state.ensureSameThread()
		val (v0, v1, v2) = value
		glProgramUniform3f(program.pointer, pointer, v0, v1, v2)
		return this
	}

	fun set(value: Int): GLUniformLocation {
		program.state.ensureSameThread()
		glProgramUniform1i(program.pointer, pointer, value)
		return this
	}

	fun set(value: FloatMatrix<*>): GLUniformLocation {
		program.state.ensureSameThread()

		if (value.rows == 3 && value.columns == 3) {
			// Матрица 3x3
			if (value is AbstractMatrix3f) {
				glProgramUniformMatrix3fv(program.pointer, pointer, false, value.toFloatBuffer())
			} else {
				glProgramUniformMatrix3fv(program.pointer, pointer, false, value.toFloatArray())
			}

			checkForGLError()
		} else if (value.rows == 4 && value.columns == 4) {
			// Матрица 4x4
			if (value is AbstractMatrix4f) {
				glProgramUniformMatrix4fv(program.pointer, pointer, false, value.toFloatBuffer())
			} else {
				glProgramUniformMatrix4fv(program.pointer, pointer, false, value.toFloatArray())
			}

			checkForGLError()
		} else {
			throw IllegalArgumentException("Can not use matrix with these dimensions: ${value.rows}x${value.columns}")
		}

		return this
	}
}

open class GLShaderProgram(val state: GLStateTracker, vararg shaders: GLShader) {
	val pointer = glCreateProgram()
	var linked = false
		private set

	private val attached = HashSet<GLShader>()
	val access = object : Collection<GLShader> by attached {}

	private val locationCache = Object2ObjectArrayMap<String, Optional<GLUniformLocation>>()

	/**
	 * Возвращает GLUniformLocation или null, если у данной программы нет такого uniform
	 *
	 * В т.ч. если она была в коде шейдера, но нигде не использовалась (отсечена компилятором)
	 *
	 * Результат поиска кешируется, но для повышения производительности вызывающий код желательно так же
	 * должен кешировать результат (как локальное свойство или переменная, в случае многократного использования)
	 */
	operator fun get(name: String): GLUniformLocation? {
		state.ensureSameThread()
		check(linked) { "Shader program is not linked!" }

		return locationCache.computeIfAbsent(name, Object2ObjectFunction {
			val location = glGetUniformLocation(pointer, name)

			if (location == -1)
				return@Object2ObjectFunction Optional.empty<GLUniformLocation>()

			return@Object2ObjectFunction Optional.of(GLUniformLocation(this, name, location))
		}).orElse(null)
	}

	operator fun set(name: String, value: IStruct4f) = this[name]?.set(value)
	operator fun set(name: String, value: IStruct3f) = this[name]?.set(value)
	operator fun set(name: String, value: Int) = this[name]?.set(value)
	operator fun set(name: String, value: FloatMatrix<*>) = this[name]?.set(value)

	fun attach(shader: GLShader) {
		state.ensureSameThread()
		check(!linked) { "Program is already linked!" }

		if (!attached.add(shader)) {
			throw IllegalStateException("Already attached! $shader")
		}
		require(!shader.unlinked) { "$shader is already unlinked, and thus can not be used" }

		glAttachShader(pointer, shader.pointer)
	}

	fun link() {
		check(!linked) { "Program is already linked!" }
		glLinkProgram(pointer)

		val success = intArrayOf(0)
		glGetProgramiv(pointer, GL_LINK_STATUS, success)

		if (success[0] == 0) {
			throw ShaderLinkException(glGetShaderInfoLog(pointer))
		}

		glGetError()

		linked = true
	}

	fun use() = state.use(this)

	fun unlinkChildren() {
		attached.forEach(GLShader::unlink)
	}

	init {
		shaders.forEach(this::attach)
	}
}
