package ru.dbotthepony.kstarbound.client.gl

import org.lwjgl.opengl.GL46.*
import ru.dbotthepony.kstarbound.math.AABB
import java.io.Closeable
import java.nio.ByteBuffer
import java.nio.ByteOrder
import kotlin.collections.ArrayList
import kotlin.math.cos
import kotlin.math.sin

enum class VertexType(val elements: Int, val indicies: IntArray) {
	LINES(2, intArrayOf(0, 1)),
	TRIANGLES(3, intArrayOf(0, 1, 2)),
	QUADS(4, intArrayOf(0, 1, 2, 1, 2, 3)),
	QUADS_AS_LINES(4, intArrayOf(0, 1, 0, 2, 1, 3, 2, 3)),
	QUADS_AS_LINES_WIREFRAME(4, intArrayOf(0, 1, 0, 2, 1, 3, 2, 3, 0, 3, 1, 2)),
}

interface IVertexBuilder<This : IVertexBuilder<This, VertexType>, VertexType : IVertex<VertexType, This>> {
	val type: ru.dbotthepony.kstarbound.client.gl.VertexType
	val indexCount: Int

	fun begin(): This
	fun vertex(): VertexType
	fun checkValid()
	fun upload(vbo: GLVertexBufferObject, ebo: GLVertexBufferObject, drawType: Int = GL_DYNAMIC_DRAW)

	fun quad(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		lambda: VertexTransformer = emptyTransform
	): This {
		check(type.elements == 4) { "Currently building $type" }

		lambda(vertex().pushVec2f(x0, y0), 0).end()
		lambda(vertex().pushVec2f(x1, y0), 1).end()
		lambda(vertex().pushVec2f(x0, y1), 2).end()
		lambda(vertex().pushVec2f(x1, y1), 3).end()

		return this as This
	}

	fun quadRotated(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		x: Float,
		y: Float,
		angle: Double,
		lambda: VertexTransformer = emptyTransform
	): This {
		check(type.elements == 4) { "Currently building $type" }

		val s = sin(angle).toFloat()
		val c = cos(angle).toFloat()

		lambda(vertex().pushVec2f(x + x0 * c - s * y0, y + s * x0 + c * y0), 0).end()
		lambda(vertex().pushVec2f(x + x1 * c - s * y0, y + s * x1 + c * y0), 1).end()
		lambda(vertex().pushVec2f(x + x0 * c - s * y1, y + s * x0 + c * y1), 2).end()
		lambda(vertex().pushVec2f(x + x1 * c - s * y1, y + s * x1 + c * y1), 3).end()

		return this as This
	}

	fun quad(aabb: AABB, lambda: VertexTransformer = emptyTransform): This {
		return quad(
			aabb.mins.x.toFloat(),
			aabb.mins.y.toFloat(),
			aabb.maxs.x.toFloat(),
			aabb.maxs.y.toFloat(),
			lambda
		)
	}

	fun quadZ(
		x0: Float,
		y0: Float,
		x1: Float,
		y1: Float,
		z: Float,
		lambda: VertexTransformer = emptyTransform
	): This {
		check(type.elements == 4) { "Currently building $type" }

		lambda(vertex().pushVec3f(x0, y0, z), 0).end()
		lambda(vertex().pushVec3f(x1, y0, z), 1).end()
		lambda(vertex().pushVec3f(x0, y1, z), 2).end()
		lambda(vertex().pushVec3f(x1, y1, z), 3).end()

		return this as This
	}
}

interface IVertex<This : IVertex<This, VertexBuilderType>, VertexBuilderType> {
	fun checkValid()
	fun expect(name: String): This
	fun expect(type: GLType): This
	fun pushVec3f(x: Float, y: Float, z: Float): This
	fun pushVec2f(x: Float, y: Float): This
	fun end(): VertexBuilderType
}

typealias VertexTransformer = (IVertex<*, *>, Int) -> IVertex<*, *>
private val emptyTransform: VertexTransformer = { it, _ -> it }
private val EMPTY_BUFFER = ByteBuffer.allocateDirect(0)

object VertexTransformers {
	fun uv(u0: Float,
		   v0: Float,
		   u1: Float,
		   v1: Float,
		   lambda: VertexTransformer = emptyTransform
	): VertexTransformer {
		return transformer@{ it, index ->
			when (index) {
				0 -> it.pushVec2f(u0, v0)
				1 -> it.pushVec2f(u1, v0)
				2 -> it.pushVec2f(u0, v1)
				3 -> it.pushVec2f(u1, v1)
			}

			return@transformer lambda(it, index)
		}
	}

	fun uv(): VertexTransformer {
		return transformer@{ it, index ->
			when (index) {
				0 -> it.pushVec2f(0f, 0f)
				1 -> it.pushVec2f(1f, 0f)
				2 -> it.pushVec2f(0f, 1f)
				3 -> it.pushVec2f(1f, 1f)
			}

			return@transformer it
		}
	}

	fun uv(lambda: VertexTransformer): VertexTransformer {
		return transformer@{ it, index ->
			when (index) {
				0 -> it.pushVec2f(0f, 0f)
				1 -> it.pushVec2f(1f, 0f)
				2 -> it.pushVec2f(0f, 1f)
				3 -> it.pushVec2f(1f, 1f)
			}

			return@transformer lambda(it, index)
		}
	}
}

class DynamicVertexBuilder(val attributes: GLFlatAttributeList, override val type: VertexType) : IVertexBuilder<DynamicVertexBuilder, DynamicVertexBuilder.Vertex> {
	private val verticies = ArrayList<Vertex>()
	override val indexCount get() = (verticies.size / type.elements) * type.indicies.size

	override fun begin(): DynamicVertexBuilder {
		verticies.clear()
		return this
	}

	override fun vertex(): Vertex {
		return Vertex()
	}

	override fun checkValid() {
		for (vertex in verticies) {
			vertex.checkValid()
		}
	}

	/**
	 * Загружает (копирует) данные в указанные буферы, с их текущей позиции
	 */
	fun upload(
		vertexBuffer: ByteBuffer,
		elementBuffer: ByteBuffer,
	) {
		check(verticies.size % type.elements == 0) { "Not fully built (expected ${type.elements} verticies to be present for each element, last element has only ${verticies.size % type.elements})" }

		require(vertexBuffer.order() == ByteOrder.nativeOrder()) { "Byte order of $vertexBuffer does not match native order" }
		require(elementBuffer.order() == ByteOrder.nativeOrder()) { "Byte order of $elementBuffer does not match native order" }

		checkValid()

		for (vertex in verticies) {
			vertex.upload(vertexBuffer)
		}

		var offsetVertex = 0

		for (i in 0 until verticies.size / type.elements) {
			for (i2 in type.indicies.indices) {
				elementBuffer.putInt(type.indicies[i2] + offsetVertex)
			}

			offsetVertex += type.elements
		}
	}

	/**
	 * Загружает буфер в указанные VBO и EBO
	 *
	 * операция создаёт мусор вне кучи и довольно медленная
	 */
	override fun upload(vbo: GLVertexBufferObject, ebo: GLVertexBufferObject, drawType: Int) {
		require(vbo.isArray) { "$vbo is not an array" }
		require(ebo.isElementArray) { "$vbo is not an element array" }

		check(verticies.size % type.elements == 0) { "Not fully built (expected ${type.elements} verticies to be present for each element, last element has only ${verticies.size % type.elements})" }

		checkValid()

		if (verticies.size == 0) {
			vbo.bufferData(EMPTY_BUFFER, drawType)
			ebo.bufferData(EMPTY_BUFFER, drawType)

			return
		}

		val vertexBuffer = ByteBuffer.allocateDirect(verticies.size * attributes.stride)
		vertexBuffer.order(ByteOrder.nativeOrder())

		val elementBuffer = ByteBuffer.allocateDirect((verticies.size / type.elements) * type.indicies.size * 4)
		elementBuffer.order(ByteOrder.nativeOrder())

		upload(vertexBuffer, elementBuffer)

		check(vertexBuffer.position() == vertexBuffer.capacity()) { "Vertex Buffer is not fully filled (position: ${vertexBuffer.position()}; capacity: ${vertexBuffer.capacity()})" }
		check(elementBuffer.position() == elementBuffer.capacity()) { "Element Buffer is not fully filled (position: ${elementBuffer.position()}; capacity: ${elementBuffer.capacity()})" }

		vertexBuffer.position(0)
		elementBuffer.position(0)

		vbo.bufferData(vertexBuffer, drawType)
		ebo.bufferData(elementBuffer, drawType)
	}

	inner class Vertex : IVertex<Vertex, DynamicVertexBuilder> {
		init {
			verticies.add(this)
		}

		private val store = arrayOfNulls<Any>(attributes.size)
		private var index = 0

		fun upload(bytes: ByteBuffer) {
			for (element in store) {
				when (element) {
					is FloatArray -> for (i in element) bytes.putFloat(i)
					is IntArray -> for (i in element) bytes.putInt(i)
					is ByteArray -> for (i in element) bytes.put(i)
					is DoubleArray -> for (i in element) bytes.putDouble(i)
					else -> throw IllegalStateException("Unknown element $element")
				}
			}
		}

		override fun toString(): String {
			return "Vertex(${store.map {
				return@map when (it) {
					is FloatArray -> it.joinToString(", ")
					is IntArray -> it.joinToString(", ")
					is ByteArray -> it.joinToString(", ")
					is DoubleArray -> it.joinToString(", ")
					else -> "null"
				} }.joinToString("; ")})"
		}

		override fun expect(name: String): Vertex {
			if (index >= attributes.size) {
				throw IllegalStateException("Reached end of attribute list early, expected $name")
			}

			if (attributes[index].name != name) {
				throw IllegalStateException("Expected $name, got ${attributes[index].name}[${attributes[index].glType}] (at position $index)")
			}

			return this
		}

		override fun expect(type: GLType): Vertex {
			if (index >= attributes.size) {
				throw IllegalStateException("Reached end of attribute list early, expected type $type")
			}

			if (attributes[index].glType != type) {
				throw IllegalStateException("Expected $type, got ${attributes[index].name}[${attributes[index].glType}] (at position $index)")
			}

			return this
		}

		override fun pushVec3f(x: Float, y: Float, z: Float): Vertex {
			expect(GLType.VEC3F)
			store[index++] = floatArrayOf(x, y, z)
			return this
		}

		override fun pushVec2f(x: Float, y: Float): Vertex {
			expect(GLType.VEC2F)
			store[index++] = floatArrayOf(x, y)
			return this
		}

		override fun checkValid() {
			for (elem in store.indices) {
				if (store[elem] == null) {
					throw IllegalStateException("Vertex element at position $elem is null")
				}
			}
		}

		override fun end(): DynamicVertexBuilder {
			checkValid()
			return this@DynamicVertexBuilder
		}
	}
}

/**
 * "Поточная" версия [DynamicVertexBuilder], ориентированная на скорость работы и имеющая фиксированный размер буфера
 *
 * Главные отличия:
 * * Данный объект не желательно создавать каждый раз когда надо отрисовать какое либо количество геометрии, а использовать уже существующий, который
 * удовлетворяет требованиям (формат вершин, и их потенциально максимальное количество)
 * * Максимальное количество vertex'ов фиксированно и равняется [maxElements] * [VertexType.elements]
 * * Имеет два встроенных DirectByteBuffer и НЕ позволяет загружать данные в другие ByteBuffer, только во внутренние ByteBuffer и только в VBO; EBO
 */
class StreamVertexBuilder(
	val attributes: GLFlatAttributeList,
	override val type: VertexType,
	val maxElements: Int,
) : IVertexBuilder<StreamVertexBuilder, StreamVertexBuilder.Vertex> {
	val maxVertexNum = maxElements * type.elements
	var nextVertex = 0
		private set

	override val indexCount get() = (nextVertex / type.elements) * type.indicies.size
	val maxIndexCount = maxElements * type.indicies.size

	val elementIndexType = when (maxIndexCount) {
		// api performance issue 102: glDrawElements uses element index type 'GL_UNSIGNED_BYTE' that is not optimal for the current hardware configuration; consider using 'GL_UNSIGNED_SHORT' instead
		// in 0 .. 255 -> GL_UNSIGNED_BYTE
		in 0 .. 65535 -> GL_UNSIGNED_SHORT
		else -> GL_UNSIGNED_INT
	}

	private var head: Vertex? = null

	/**
	 * Буфер для VBO достаточного размера
	 */
	private val vertexBuffer = ByteBuffer.allocateDirect(maxVertexNum * attributes.stride)

	/**
	 * Буфер для EBO достаточного размера
	 */
	private val elementBuffer = ByteBuffer.allocateDirect(maxElements * type.indicies.size * 4)

	init {
		vertexBuffer.order(ByteOrder.nativeOrder())
		elementBuffer.order(ByteOrder.nativeOrder())
	}

	private fun writeElementIndex(value: Int) {
		when (elementIndexType) {
			GL_UNSIGNED_BYTE -> elementBuffer.put(value.toByte())
			GL_UNSIGNED_SHORT -> elementBuffer.putShort(value.toShort())
			else -> elementBuffer.putInt(value)
		}
	}

	private var offsetElementIndex = 0

	/**
	 * Устанавливает метку этого билдера в ноль.
	 *
	 * Не обнуляет память буферов!
	 */
	override fun begin(): StreamVertexBuilder {
		nextVertex = 0
		offsetElementIndex = 0
		head = null
		vertexBuffer.position(0)
		elementBuffer.position(0)
		return this
	}

	override fun vertex(): Vertex {
		return Vertex()
	}

	override fun upload(vbo: GLVertexBufferObject, ebo: GLVertexBufferObject, drawType: Int) {
		require(vbo.isArray) { "$vbo is not an array" }
		require(ebo.isElementArray) { "$vbo is not an element array" }

		if (nextVertex == 0) {
			vbo.bufferData(EMPTY_BUFFER, drawType)
			ebo.bufferData(EMPTY_BUFFER, drawType)

			return
		}

		checkValid()

		val a = vertexBuffer.position().toLong()
		val b = elementBuffer.position().toLong()

		vertexBuffer.position(0)
		elementBuffer.position(0)

		vbo.bufferData(vertexBuffer, drawType, length = a)
		ebo.bufferData(elementBuffer, drawType, length = b)
	}

	override fun checkValid() {
		var vertex = head

		while (vertex != null) {
			vertex.checkValid()
			vertex = vertex.previous
		}
	}

	inner class Vertex : IVertex<Vertex, StreamVertexBuilder> {
		private val vertexIndex = nextVertex++
		val previous = head
		private var bufferPosition = vertexIndex * attributes.stride

		init {
			if (vertexIndex >= maxVertexNum) {
				throw IndexOutOfBoundsException("Tried to push new vertex $vertexIndex, when already above limit of $maxVertexNum!")
			}

			head = this

			for (i2 in type.indicies.indices) {
				writeElementIndex(type.indicies[i2] + offsetElementIndex)
			}

			offsetElementIndex += type.elements
		}

		private var index = 0

		override fun checkValid() {
			check(index == attributes.size) { "Vertex $vertexIndex is not fully filled (only $index attributes provided, ${attributes.size} required)" }
		}

		override fun expect(name: String): Vertex {
			if (index >= attributes.size) {
				throw IllegalStateException("Reached end of attribute list early, expected $name")
			}

			if (attributes[index].name != name) {
				throw IllegalStateException("Expected $name, got ${attributes[index].name}[${attributes[index].glType}] (at position $index)")
			}

			return this
		}

		override fun expect(type: GLType): Vertex {
			if (index >= attributes.size) {
				throw IllegalStateException("Reached end of attribute list early, expected type $type")
			}

			if (attributes[index].glType != type) {
				throw IllegalStateException("Expected $type, got ${attributes[index].name}[${attributes[index].glType}] (at position $index)")
			}

			return this
		}

		override fun pushVec3f(x: Float, y: Float, z: Float): Vertex {
			expect(GLType.VEC3F)
			vertexBuffer.position(bufferPosition)
			vertexBuffer.putFloat(x)
			vertexBuffer.putFloat(y)
			vertexBuffer.putFloat(z)
			index++
			bufferPosition += 12
			return this
		}

		override fun pushVec2f(x: Float, y: Float): Vertex {
			expect(GLType.VEC2F)
			vertexBuffer.position(bufferPosition)
			vertexBuffer.putFloat(x)
			vertexBuffer.putFloat(y)
			index++
			bufferPosition += 8
			return this
		}

		override fun end(): StreamVertexBuilder {
			check(index == attributes.size) { "Vertex $vertexIndex is not fully filled (only $index attributes provided, ${attributes.size} required)" }
			return this@StreamVertexBuilder
		}
	}
}

class StatefulStreamVertexBuilder(
	val state: GLStateTracker,
	val builder: StreamVertexBuilder
) : Closeable, IVertexBuilder<StreamVertexBuilder, StreamVertexBuilder.Vertex> by builder {
	private val vao = state.newVAO()
	private val vbo = state.newVBO()
	private val ebo = state.newEBO()

	init {
		vao.bind()
		vbo.bind()
		ebo.bind()

		builder.attributes.apply(vao, true)

		vao.unbind()
		vbo.unbind()
		ebo.unbind()
	}

	fun upload(drawType: Int = GL_DYNAMIC_DRAW) {
		builder.upload(vbo, ebo, drawType)
	}

	fun bind() = vao.bind()
	fun unbind() = vao.unbind()

	fun draw(primitives: Int = GL_TRIANGLES) {
		bind()
		glDrawElements(primitives, builder.indexCount, builder.elementIndexType, 0L)
		checkForGLError()
	}

	override fun close() {
		vao.close()
		vbo.close()
		ebo.close()
	}
}
