package ru.dbotthepony.kstarbound.util

import com.google.common.collect.ImmutableList
import com.google.gson.*
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.api.IStruct4f
import java.lang.reflect.Type

data class Color(val red: Float, val green: Float, val blue: Float, val alpha: Float = 1f) : IStruct4f {
	constructor(input: JsonArray) : this(input[0].asFloat / 255f, input[1].asFloat / 255f, input[2].asFloat / 255f, if (input.size() >= 4) input[3].asFloat / 255f else 1f)

	constructor(input: Long) : this(
		((input ushr 16) and 0xFFL).toFloat() / 255f,
		((input ushr 8) and 0xFFL).toFloat() / 255f,
		(input and 0xFFL).toFloat() / 255f,
	)

	companion object {
		val WHITE = Color(1f, 1f, 1f)

		val RED = Color(1f, 0f, 0f)
		val GREEN = Color(0f, 1f, 0f)
		val BLUE = Color(0f, 0f, 1f)

		val SLATE_GREY = Color(0.2f, 0.2f, 0.2f)

		val PRE_DEFINED_MAP = mapOf(
			"red" to RED,
			"green" to GREEN,
			"blue" to BLUE,
		)

		val SHADES_OF_GRAY = ArrayList<Color>().let {
			for (i in 0 .. 256) {
				it.add(Color(i / 256f, i / 256f, i / 256f))
			}

			return@let ImmutableList.copyOf(it) as List<Color>
		}
	}
}

object ColorTypeAdapter : TypeAdapter<Color>() {
	override fun write(out: JsonWriter, value: Color) {
		TODO("Not yet implemented")
	}

	override fun read(`in`: JsonReader): Color {
		when (val type = `in`.peek()) {
			JsonToken.BEGIN_ARRAY -> {
				`in`.beginArray()
				val red = `in`.nextDouble()
				val green = `in`.nextDouble()
				val blue = `in`.nextDouble()

				if (red % 1.0 == 0.0 && green % 1.0 == 0.0 && blue % 1.0 == 0.0) {
					val alpha = `in`.peek().let { if (it == JsonToken.END_ARRAY) 255.0 else `in`.nextDouble() }
					`in`.endArray()

					return Color(
						red.toFloat() / 255f,
						green.toFloat() / 255f,
						blue.toFloat() / 255f,
						alpha.toFloat() / 255f,
					)
				} else {
					val alpha = `in`.peek().let { if (it == JsonToken.END_ARRAY) 1.0 else `in`.nextDouble() }
					`in`.endArray()

					return Color(
						red.toFloat(),
						green.toFloat(),
						blue.toFloat(),
						alpha.toFloat(),
					)
				}
			}

			JsonToken.BEGIN_OBJECT -> {
				`in`.beginObject()

				val keyed = mutableMapOf<String, Double>()

				while (`in`.peek() != JsonToken.END_OBJECT) {
					keyed[`in`.nextName()] = `in`.nextDouble()
				}

				if (keyed.isEmpty())
					throw IllegalArgumentException("Object is empty")

				var values = 0

				val red = keyed["red"]?.also { values++ } ?: keyed["r"]?.also { values++ } ?: 255.0
				val green = keyed["green"]?.also { values++ } ?: keyed["g"]?.also { values++ } ?: 255.0
				val blue = keyed["blue"]?.also { values++ } ?: keyed["b"]?.also { values++ } ?: 255.0
				val alpha = keyed["alpha"]?.also { values++ } ?: keyed["a"]?.also { values++ } ?: 255.0
				`in`.endObject()

				if (values == 0) {
					throw IllegalArgumentException("Object is not a color")
				}

				if (red % 1.0 == 0.0 && green % 1.0 == 0.0 && blue % 1.0 == 0.0 && alpha % 1.0 == 0.0) {
					return Color(
						red.toFloat() / 255f,
						green.toFloat() / 255f,
						blue.toFloat() / 255f,
						alpha.toFloat() / 255f,
					)
				} else {
					return Color(
						red.toFloat(),
						green.toFloat(),
						blue.toFloat(),
						alpha.toFloat(),
					)
				}
			}

			JsonToken.NUMBER -> return Color(`in`.nextLong())
			JsonToken.STRING -> {
				val name = `in`.nextString()
				return Color.PRE_DEFINED_MAP[name] ?: throw IllegalArgumentException("Unknown pre defined color name $name")
			}

			else -> throw IllegalArgumentException("Expected array, object or number; got $type")
		}
	}
}
